import {
  Component,
  ElementRef,
  ViewChild,
  OnInit,
  Input,
  OnDestroy,
  HostListener
  
} from "@angular/core";

import {
  MySprite,
  RichText,
  getMinScale,
  ShapeRect,
  ShapeCircle,
  tweenChange,
  randomSortByArr,
  Label,
  showPopParticle,
  moveItem,
  removeItemFromArr,
  rotateItem,
  hideItem,
  showItem,
  ShapeRectNew,
  scaleItem,
  showBlingBling,
  waterWave,
  jelly,
  getAngleByPos,
  getPosDistance,
  shake
} from "./Unit";

import { localImages, localAudios, multiSizeBackground} from "./resources";

import { Cartoon } from './Cartoon'
import { Subject } from "rxjs";
import { debounceTime, map, takeWhile, retry } from "rxjs/operators";
import * as _ from "lodash";
import TWEEN from "@tweenjs/tween.js";
import { demoData as defaultData } from '../../assets/play/data/data.js'
import { v4 as uuidv4 } from 'uuid';

const zIndexMap = {
  mainBackground: 0,
  sliderBG: 100,
  title: 100,
  buttons: 100,
  topMaskLine: 99
}

@Component({
  selector: "app-play",
  templateUrl: "./play.component.html",
  styleUrls: ["./play.component.scss"]
})

export class PlayComponent implements OnInit, OnDestroy {
  
  g_cartoon = new Cartoon()

  // ------------  全局数据  ------------
  g_topElementContainer; // 用于装载装饰元素 气球 星星等
  g_stage; //中心舞台
  // g_background_color = "#rgb(0,0,0,0)"
  g_background_color = "#f0f5f8"
  g_enableMapDown = true; // 触摸使能
  g_enableMapUp = true; // 抬起使能
  g_enableMapMove = true; // 移动ss使能
  g_canvasLeft;
  g_canvasTop;
  g_animationId: any;
  g_mapScale = 1; // 缩放比例
  g_KEY = "DataKey_YM5-9";
  g_canvasWidth = 1280;
  g_canvasHeight = 720;
  g_canvasBaseW = 1280;
  g_canvasBaseH = 720;
  g_winResizeEventStream = new Subject();
  g_clickX; // 点击坐标 X
  g_clickY; // 点击坐标 Y
  g_ctx; // canvas 实例
  g_data; // 数据
  g_formData; // 核心表单数据
  g_teacherFlag = false; // 默认角色
  g_currentUser;
  g_firstTouch = true;
  g_partTitle_x = null;
  g_mainTitle_x = null;
  // ------------------------------------

  // ------------  私有数据  ------------
  m_mapDownQueue = {} //按下事件处理队列
  m_mapDownArray = [] //按下事件处理队列
  m_mapDownObject = []

  m_mapUpQueue = {} //抬起事件处理队列
  m_mapUpArray = [] //抬起事件处理队列
  m_mapUpObject = []

  m_mapDragQueue = {} //拖拽事件处理队列
  m_mapDragArray = [] //拖拽事件处理队列
  m_mapDragObject = []

  m_mapMoveArray = [] //移动事件处理队列
  m_mapMoveObject = []
  m_endPageArr;
  m_showPetalFlag;
  m_elementPetalArr;
  m_showElementPetalFlag;
  m_PetalImage = "_scrap-pic-"  // 飘落动画
  m_renderArr // 渲染队列
  m_renderObject = [];
  m_frontCurtainArr = []; // 前置幕布
  m_defaultZindex = 0;
  m_setTimeoutIDs = [];
  m_setIntervalIDs = [];
  m_moveAsstantIntervalId = null;

  g_currentDragElementID: any = null; // 当前拖动的动画元件
  g_dragFirstClickY: any;
  g_dragFirstClickX: any;
  g_enableDragDistance: number = 20 * this.g_mapScale;
  g_enableDragging: any;
  g_maskLayerZindex = 100;
  // ------------------------------------
  

  // ------------  游戏逻辑数据  ------------
  m_currentQuestionIndex: number = -1;
  // ------------------------------------


  // ------------  消息  ------------
  
  // ------------------------------------


  // ------------  调试变量  ------------
  g_EnableStageRuler = false; // 使能舞台背景格尺
  g_ForceChangeDefaultRole = false // 强制当前角色为默认角色
  g_EnableTestSendEvent = false // 发送模拟Web数据
  g_showLeftCornerTest= false // 测试左上角图标
  // ------------------------------------


  // 当数据加载完毕后，执行
  systemReady(){
    this.setLeftCornerTest()
    this.initGame()
  }
  
  // 屏幕尺寸变化后执行
  handleScreenResize(){
    this.cleanSystemVar()
    this.cleanGameVar();
    this.setLeftCornerTest();
    this.disableMoveAsstant();
    this.initSystem();
    this.initGame()
  }

  // 映射预加载图片[网路]资源 返回包含图片路径的数组
  mapToImageArray(contentObj){
    let array = []

    return array
  }

  // 映射预加载音频[网路]资源 返回包含音频路径的数组
  mapToAduioArray(contentObj){
    let array = []
    contentObj.title.mainTextAudio_url && array.push(contentObj.title.mainTextAudio_url);
    contentObj.dataArray.forEach(element => {
      array.push(element.audio_url_answer);
      array.push(element.audio_url_question);
    });
    return array
  }

  // ------------------------------------------------------------------------------
  // 游戏核心处理区
  // ------------------------------------------------------------------------------
  // ------------------------------------------------------------------------------
  // 
  // 
  // 
  // ------------------------------------------------------------------------------
  // ------------------------------------------------------------------------------

  initGame(){
    console.log(this.g_formData);
    this.createTestLabel();
    this.initTitle_Part();
    this.initTitle_Main();
    
    this.initScrollPannel();
    this.initSlider();
    this.initButtions();
    this.initTopMaskLine();

    this.nextQuestion()
    this.g_cartoon.playAudio("open")
  }

  cleanGameVar(){
    this.m_currentQuestionIndex = -1;
  }

  startGame(){

  }

  restartGame(){
    // this.cleanGameVar()
    this.startGame()
  }

  initTitle_Part(){
    let element =  this.g_cartoon.createCartoonElementImageFunc("part-title", "bg_title_part", (w, h)=>{
      return {
        sx: 57*this.g_mapScale / w,
        sy: 65*this.g_mapScale / h,
      }
    }, (w, h)=>{
      return {
        x: 53*this.g_mapScale + (57*this.g_mapScale)/2,
        y: (65*this.g_mapScale)/2
      }
    })
    element.ref.x = this.g_partTitle_x - 20 * this.g_mapScale + (57*this.g_mapScale)/2;
    let text = this.g_cartoon.createLabel(this.g_formData.title.NO?this.g_formData.title.NO:"C", "BRLNSDB_1", "#FFFFFF", 48)
    element.ref.addChild(text)

    let partBG = this.g_cartoon.createCartoonElementImageFunc("part-title-background", "part_number_bg", (w, h)=>{
      return {
        sx: 121*this.g_mapScale / w,
        sy: 69*this.g_mapScale / h,
      }
    }, (w, h)=>{
      return {
        x: (121*this.g_mapScale)/2,
        y: (69*this.g_mapScale)/2
      }
    })

    this.render(partBG.ref, zIndexMap.title)
    this.render(element.ref, zIndexMap.title)
  }

  initTitle_Main(){
    let labelWidth = 0
    let text =  this.g_cartoon.createCartoonElementLabelFunc("main-title", this.g_formData.title.mainText, "FuturaStd-Medium", "#000000", 36, (w, h)=>{
      labelWidth = w/2
      return {
        x: this.g_partTitle_x - 20 * this.g_mapScale + labelWidth * this.g_mapScale,
        y: 5*this.g_mapScale + (65*this.g_mapScale) / 2
      }
    })
    text.ref.fontWeight = 900
    text.ref.setScaleXY(this.g_mapScale)

    let background = this.g_cartoon.createCartoonElementImageFunc("main-title-background", "_bg_50_20", (w, h)=>{
      return {
        sx: labelWidth*2 / w,
        sy: 50*this.g_mapScale / h,
      }
    }, (w, h)=>{
      return {
        x: this.g_partTitle_x + labelWidth  + 60 * this.g_mapScale,
        y: 5*this.g_mapScale + (65*this.g_mapScale) / 2
      }
    })

    text.ref.x = this.g_partTitle_x + 60 * this.g_mapScale
    text.ref.textAlign = "left"

    this.subscribeMapDownEvent(background.id, ()=>{
      this.g_cartoon.stopAllAudio()
      this.g_cartoon.playAudio(this.g_formData.title.mainTextAudio_url);
      this.g_enableMapDown = true;
    })

    this.render(background.ref, zIndexMap.title)
    this.render(text.ref, zIndexMap.title)
  }

  initTopMaskLine() {
    let maskTop = this.g_cartoon.createCartoonElementShapeRect("top-mask", this.g_canvasWidth, 80 * this.g_mapScale, "#f0f5f8") // 
    this.render(maskTop.ref, zIndexMap.topMaskLine)
  }

  initSlider() {
    const updateSlider = (len) => {
      let top = this.g_cartoon.getCartoonElement("slider-top");
      let body = this.g_cartoon.getCartoonElement("slider-body");
      let bottom = this.g_cartoon.getCartoonElement("slider-bottom");
      body.update(len);
      top.update();
      bottom.update();
    }

    const setProgress = (scale) => {
      sliderBG.currentProgress = scale;
      let scrollPannel = this.g_cartoon.getCartoonElement("scroll-pannel")
      let totalHeight = 546;
      let pix = totalHeight / sliderBG.fullHeight;
      let empty = (sliderBG.fullHeight - sliderBG.winHeight) * scale * pix;
      let top = this.g_cartoon.getCartoonElement("slider-top");
      let body = this.g_cartoon.getCartoonElement("slider-body");
      let bottom = this.g_cartoon.getCartoonElement("slider-bottom");
      let bb = body.ref.getBoundingBox()
      body.ref.y = -282 + empty + 4 + bb.height / 2 / this.g_mapScale;

      top.update();
      bottom.update();

      scrollPannel.ref.y = 78 * this.g_mapScale - (scrollPannel.currentY - (this.g_canvasHeight - 78 * this.g_mapScale)) * sliderBG.currentProgress;
    }

    let sliderBG = this.g_cartoon.createCartoonElementImageFunc("slider-background", "bgSlider", (w, h)=>{
      return {
        sx: 14 * this.g_mapScale / w,
        sy: 552 * this.g_mapScale / h,
      }
    }, (w, h)=>{
      return {
        x: this.g_canvasWidth - 110 * this.g_mapScale,
        y: this.g_canvasHeight / 2
      }
    })
    sliderBG.ref.scaleX = 0;
    sliderBG.isShow = false;
    sliderBG.show = () => {
      if(!sliderBG.isShow) {
        sliderBG.isShow = true;
        tweenChange(sliderBG.ref, {scaleX: sliderBG.initScaleX}, 0.05)
      }
    }
    sliderBG.hide = () => {
      tweenChange(sliderBG.ref, {scaleX: 0},  0.05)
    }
    sliderBG.initSlider = (windowHeight, fullHeight, current) => {
      sliderBG.winHeight = windowHeight;
      sliderBG.fullHeight = fullHeight;
      let totalHeight = 546;
      let pix = totalHeight / fullHeight;
      let fillHeight = windowHeight * pix

      updateSlider(fillHeight)
      setProgress(current)
    }
    sliderBG.winHeight = 0;
    sliderBG.fullHeight = 0;
    sliderBG.pix = 0;
    sliderBG.currentProgress = 0;

    sliderBG.getCurrentProgress = () => {
      let emptyTop = (sliderTop.ref.y - 2) + 271;
      let emptyDown = 271 - (sliderBottom.ref.y + 2)
      let scale = emptyTop / (emptyTop + emptyDown)
      let progress = Math.floor(scale * 100) / 100
      return progress>0?progress:0
    }

    let sliderTop = this.g_cartoon.createCartoonElementImageFunc("slider-top", "slider_top", (w, h)=>{
      return {
        sx: 8 / w,
        sy: 4 / h,
      }
    }, (w, h)=>{
      return {
        x: 0,
        y: -271
      }
    })
    sliderTop.update = () => {
      let y = sliderBody.ref.y
      let box = sliderBody.ref.getBoundingBox()
      y -= box.height / 2 / this.g_mapScale + 2;
      if(y<-271) {
        // console.log("滑动已到最上", y)
        return false;
      } else {
        sliderTop.ref.y = y;
        return true;
      }
    }

    let sliderBody = this.g_cartoon.createCartoonElementImageFunc("slider-body", "slider_body", (w, h)=>{
      return {
        sx: 8 / w,
        sy: 4 / h,
      }
    }, (w, h)=>{
      return {
        x: 0,
        y: -272
      }
    })
    sliderBody.update = (len) => {
      let y = sliderTop.ref.y + 2
      sliderBody.ref.scaleY = len / 4
      y += len / 2;
      sliderBody.ref.y = y
    }
    sliderBody.toEnd = () => {
      let box = sliderBody.ref.getBoundingBox()
      sliderBottom.ref.y = 271
      sliderBody.ref.y = sliderBottom.ref.y - box.height / 2 / this.g_mapScale - 2;
      sliderTop.update();
    }
    sliderBody.toTop = () => {
      let box = sliderBody.ref.getBoundingBox()
      sliderTop.ref.y = -271
      sliderBody.ref.y = sliderTop.ref.y + box.height / 2 / this.g_mapScale + 2;
      sliderBottom.update();
    }

    let sliderBottom = this.g_cartoon.createCartoonElementImageFunc("slider-bottom", "slider_bottom", (w, h)=>{
      return {
        sx: 8 / w,
        sy: 4 / h,
      }
    }, (w, h)=>{
      return {
        x: 0,
        y: 271
      }
    })
    sliderBottom.update = () => {
      let y = sliderBody.ref.y
      let box = sliderBody.ref.getBoundingBox()
      y += box.height / 2 / this.g_mapScale + 2;
      
      if(y>271) {
        // console.log("滑动已到最下", y)
        return false;
      } else {
        sliderBottom.ref.y = y
        return true;
      }
    }
    sliderBG.ref.addChild(sliderBody.ref)
    sliderBG.ref.addChild(sliderBottom.ref)
    sliderBG.ref.addChild(sliderTop.ref)
    this.render(sliderBG.ref, zIndexMap.sliderBG)

    let clickSY = 0
    let baseY = 0;
    const scrollPannel = this.g_cartoon.getCartoonElement("scroll-pannel")
    
    this.subscribeMapDragEvent(sliderBody.id, ()=>{
      clickSY = this.g_clickY;
      baseY = sliderBody.ref.y
    }, ()=>{
      let dis = clickSY - this.g_clickY
      let savedY = sliderBody.ref.y
      sliderBody.ref.y = baseY - dis
      let crTop = sliderTop.update()
      let crBottom = sliderBottom.update()
      if(!crTop) {
        sliderBody.toTop()
        sliderTop.update();
        sliderBottom.update();
      }
      if(!crBottom) {
        sliderBody.toEnd();
        sliderTop.update();
        sliderBottom.update();
      }
      
      scrollPannel.ref.y = 78 * this.g_mapScale - (scrollPannel.currentY - (this.g_canvasHeight - 78 * this.g_mapScale)) * sliderBG.getCurrentProgress()
    }, ()=>{
      this.g_enableMapUp = true;
      this.g_enableMapDown = true;
    })
  }

  initScrollPannel() {
    let pannel = this.g_cartoon.createCartoonElementShapeRect("scroll-pannel", this.g_canvasWidth - 180 * this.g_mapScale, this.g_canvasHeight - 52 * this.g_mapScale, "#f0f5f8") // f0f5f8
    pannel.ref.y = 52 * this.g_mapScale
    this.render(pannel.ref, 10);
    pannel.lastInserType = "Q"
    pannel.windowHeight = this.g_canvasHeight - 52 * this.g_mapScale
    pannel.currentY = 0;
    pannel.currentQuestionIndex = -1;
    pannel.topY =  52 * this.g_mapScale
    pannel.insertQuestion = (text, nextQuestion = false) => {
      if(nextQuestion) {
        pannel.currentQuestionIndex++;
      }
      this.g_cartoon.playAudio("wtcx")
      let textString = ""
      let startX = 42 * this.g_mapScale;
      if(nextQuestion) {
        if(this.g_formData.dataArray[pannel.currentQuestionIndex].showIndex) {
          textString = `${pannel.currentQuestionIndex+1}. ${text}`;
        } else {
          textString = `${text}`;
        }
        pannel.currentY += 18 * this.g_mapScale;
      } else {
        textString = text;
        pannel.currentY -= 52 * this.g_mapScale;
        startX += 42 * this.g_mapScale;
      }

      let lineText = this.creatQuestionTextLine(pannel.currentQuestionIndex, textString, startX, pannel.currentY, pannel.ref)
      let label = lineText.label;
      lineText.show()
      pannel.currentY = label.ref.y + 64 * this.g_mapScale;
      if(pannel.windowHeight < pannel.currentY) {
        pannel.ref.setSize(this.g_canvasWidth - 180 * this.g_mapScale,  pannel.currentY)
        this.g_cartoon.getCartoonElement("slider-background").initSlider(pannel.windowHeight, pannel.currentY, 1)
        this.g_cartoon.getCartoonElement("slider-background").show()
      }
      pannel.ref.addChild(label.ref);
      pannel.lastInserType = "Q";
    }
    
    pannel.canScroll = false;
    pannel.firstInstert = true;
    pannel.insertAnswer = (text) => {
      this.g_cartoon.playAudio("dacx")
      let line = this.creatAnswerTextLine(pannel.currentQuestionIndex, text, 42 * this.g_mapScale * 2, pannel.currentY, pannel.ref)
      let label = line.label
      let bb = label.ref.getBoundingBox();
      label.ref.x = 42 * this.g_mapScale * 2 + bb.width / 2;
      if(pannel.lastInserType == "Q") {
        pannel.currentY -= 32 * this.g_mapScale
        line.mask.ref.y -= 32 * this.g_mapScale;
        line.target.ref.y -= 32 * this.g_mapScale;
      }
      label.ref.y = pannel.currentY + 32 * this.g_mapScale / 2;
      
      pannel.currentY = label.ref.y + 32 * this.g_mapScale;

      line.show()
      if(pannel.windowHeight < pannel.currentY) {
        pannel.ref.setSize(this.g_canvasWidth - 180 * this.g_mapScale,  pannel.currentY)
        this.g_cartoon.getCartoonElement("slider-background").initSlider(pannel.windowHeight, pannel.currentY, 1)
        this.g_cartoon.getCartoonElement("slider-background").show();
        pannel.canScroll = true;
      }
      pannel.ref.addChild(label.ref);
      pannel.lastInserType = "A"
    }

    let clickSY = null;
    let baseY = null;
    this.subscribeMapDragEvent(pannel.id, ()=>{
      if(!pannel.canScroll) {
        this.g_enableMapUp = true;
        this.g_enableMapDown = true;
        return;
      }
      clickSY = this.g_clickY;
      baseY = pannel.ref.y;
    }, ()=>{
      let dis = clickSY - this.g_clickY;
      let _t = pannel.topY - (pannel.currentY - pannel.windowHeight)
      if(pannel.topY <= (baseY - dis)) {
        pannel.ref.y = pannel.topY 
      } else {
        if((baseY - dis) <= _t) {
          pannel.ref.y = _t
        } else {
          pannel.ref.y = baseY - dis
        }
      }
      let total = pannel.topY - _t;
      let current = pannel.topY - pannel.ref.y
      let progress = Math.floor(current / total * 100) / 100;
      if(progress<0.05) {
        progress = 0.05
      }
      if(progress>1) {
        progress = 1
      }
      this.g_cartoon.getCartoonElement("slider-background").initSlider(pannel.windowHeight, pannel.currentY, progress)
      // let crTop = sliderTop.update()
      // let crBottom = sliderBottom.update()
      // if(!crTop) {
      //   sliderBody.toTop()
      //   sliderTop.update();
      //   sliderBottom.update();
      // }
      // if(!crBottom) {
      //   sliderBody.toEnd();
      //   sliderTop.update();
      //   sliderBottom.update();
      // }
      
      // scrollPannel.ref.y = 78 * this.g_mapScale - (scrollPannel.currentY - (this.g_canvasHeight - 78 * this.g_mapScale)) * sliderBG.getCurrentProgress()
    }, ()=>{
      this.g_enableMapUp = true;
      this.g_enableMapDown = true;
    })

  }

  initButtions() {
    let btnQuestion = this.g_cartoon.createCartoonElementImageFunc("button-question", "btn_question", (w, h)=>{
      return {
        sx: 88 * this.g_mapScale / w,
        sy: 88 * this.g_mapScale / h,
      }
    }, (w, h)=>{
      return {
        x: this.g_canvasWidth - 50 * this.g_mapScale,
        y: this.g_canvasHeight / 2 - 150 * this.g_mapScale,
      }
    })
    btnQuestion.show = (callback) => {
      this.g_cartoon.playAudio("show_btn")
      btnQuestion.ref.scaleX = 0;
      btnQuestion.ref.scaleY = 0;
      tweenChange(btnQuestion.ref, {scaleX: btnQuestion.initScaleX, scaleY: btnQuestion.initScaleY}, 0.2, ()=>{
        callback && callback()
      })
    }
    btnQuestion.hide = (callback) => {
      btnQuestion.ref.scaleX = btnQuestion.initScaleX;
      btnQuestion.ref.scaleY = btnQuestion.initScaleY;
      tweenChange(btnQuestion.ref, {scaleX: 0, scaleY: 0}, 0.2, ()=>{
        callback && callback()
      })
    }

    let btnAnswer = this.g_cartoon.createCartoonElementImageFunc("button-answer", "btn_answer", (w, h)=>{
      return {
        sx: 88 * this.g_mapScale / w,
        sy: 88 * this.g_mapScale / h,
      }
    }, (w, h)=>{
      return {
        x: this.g_canvasWidth - 50 * this.g_mapScale,
        y: this.g_canvasHeight / 2 - 150 * this.g_mapScale,
      }
    })
    btnAnswer.ref.scaleX = 0;
    btnAnswer.ref.scaleY = 0;
    btnAnswer.show = (callback) => {
      this.g_cartoon.playAudio("show_btn")
      btnAnswer.ref.scaleX = 0;
      btnAnswer.ref.scaleY = 0;
      tweenChange(btnAnswer.ref, {scaleX: btnAnswer.initScaleX, scaleY: btnAnswer.initScaleY}, 0.2, ()=>{
        callback && callback()
      })
    }
    btnAnswer.hide = (callback) => {
      btnAnswer.ref.scaleX = btnAnswer.initScaleX;
      btnAnswer.ref.scaleY = btnAnswer.initScaleY;
      tweenChange(btnAnswer.ref, {scaleX: 0, scaleY: 0}, 0.2, ()=>{
        callback && callback()
      })
    }

    const scrollPannel = this.g_cartoon.getCartoonElement("scroll-pannel")
    this.subscribeMapDownEvent(btnQuestion.id, ()=>{
      let text = this.g_formData.dataArray[this.m_currentQuestionIndex].questionText?this.g_formData.dataArray[this.m_currentQuestionIndex].questionText:""
      let textArray = text.split(/\n/)
      let lastIndex = textArray.length;
      this.g_cartoon.playAudio("click_btn")
      jelly(btnQuestion.ref)
      textArray.forEach((item, index) => {
        if(index==0) {
          setTimeout(() => {
            scrollPannel.insertQuestion(item, true)
            if(index == lastIndex-1) {
              setTimeout(() => {
                btnQuestion.hide(()=>{
                  btnAnswer.show(()=>{
                    this.g_enableMapDown = true;
                  })
                })
              }, 1000);
            }
          }, 0);
        } else {
          setTimeout(() => {
            scrollPannel.insertQuestion(item, false);
            if(index == lastIndex-1) {
              setTimeout(() => {
                btnQuestion.hide(()=>{
                  btnAnswer.show(()=>{
                    this.g_enableMapDown = true;
                  })
                })
              }, 1000);
            }
          }, 900*index);
        }
      })
    }, 100)

    this.subscribeMapDownEvent(btnAnswer.id, ()=>{
      let text = this.g_formData.dataArray[this.m_currentQuestionIndex].answerText?this.g_formData.dataArray[this.m_currentQuestionIndex].answerText:""
      let textArray = text.split(/\n/)
      let lastIndex = textArray.length;
      this.g_cartoon.playAudio("click_btn")
      jelly(btnAnswer.ref);
      textArray.forEach((item, index) => {
        setTimeout(() => {
          scrollPannel.insertAnswer(item);
          if(index == lastIndex-1) {
            setTimeout(() => {
              btnAnswer.hide(()=>{
                this.nextQuestion((status)=>{
                  if(status) {
                    btnQuestion.show(()=>{
                      this.g_enableMapDown = true;
                    })
                  } else {
                    this.g_enableMapDown = true;
                  }
                })
              })
            }, 1000);
          }
        }, 900*index);
      })
    }, 100)

    this.render(btnQuestion.ref, zIndexMap.buttons);
    this.render(btnAnswer.ref,  zIndexMap.buttons);
  }

  
  creatQuestionTextLine(index, text, startX, startY, parent) {
    let line = {
      label: null,
      show: null
    }

    line.label = this.g_cartoon.createCartoonElementLabel(`lebel-question-${index}`, text, "GOTHICB_1", "#000000", 36 * this.g_mapScale);
    let bb = line.label.ref.getBoundingBox()
    line.label.ref.x = startX + bb.width / 2;
    line.label.ref.y = startY + 64 * this.g_mapScale / 2;
    

    let mask = this.g_cartoon.createCartoonElementShapeRect(`lebel-question-${index}-mask`, this.g_canvasWidth, 64 * this.g_mapScale, "#f0f5f8") // f0f5f8
    mask.ref.x = startX
    mask.ref.y = startY

    let target = this.g_cartoon.createCartoonElementShapeRect(`lebel-question-${index}-target-${uuidv4()}`, this.g_canvasWidth - 200 * this.g_mapScale, 64 * this.g_mapScale, "#FFFFFF")
    target.ref.x = startX
    target.ref.y = startY
    target.ref.alpha = 0;

    this.subscribeMapDownEvent(target.id, ()=>{
      target.audio_url = this.g_formData.dataArray[index].audio_url_question?this.g_formData.dataArray[index].audio_url_question:""
      if(target.audio_url) {
        this.g_cartoon.playAudio(target.audio_url, false, ()=>{
          this.g_enableMapDown = true;
        })
      } else {
        this.g_enableMapDown = true;
      }
    })

    line.show = () => {
      tweenChange(mask.ref, {x: this.g_canvasWidth}, 1)
    }

    parent.addChild(line.label.ref);
    parent.addChild(mask.ref);
    parent.addChild(target.ref);
    return line
  }

  creatAnswerTextLine(index, text, startX, startY, parent) {
    let line = {
      label: null,
      mask: null,
      target: null,
      show: null
    }
    line.label = this.g_cartoon.createCartoonElementLabel(`lebel-answer-${index}-${uuidv4()}`, text, "DFChu-W4-WIN-BF", "#d60d00", 42 * this.g_mapScale);
    let bb = line.label.ref.getBoundingBox()
    line.label.ref.x = startX + bb.width / 2;
    line.label.ref.y = startY + 64 * this.g_mapScale / 2;

    let mask = this.g_cartoon.createCartoonElementShapeRect(`lebel-answer-${index}-mask`, this.g_canvasWidth, 48 * this.g_mapScale, "#f0f5f8") // f0f5f8
    mask.ref.x = startX
    mask.ref.y = startY
    line.mask = mask;

    let target = this.g_cartoon.createCartoonElementShapeRect(`lebel-answer-${index}-target-${uuidv4()}`, this.g_canvasWidth - 200 * this.g_mapScale, 64 * this.g_mapScale, "#FFFFFF")
    target.ref.x = startX
    target.ref.y = startY;
    target.ref.alpha = 0;
    line.target = target;

    this.subscribeMapDownEvent(target.id, ()=>{
      target.audio_url = this.g_formData.dataArray[index].audio_url_answer?this.g_formData.dataArray[index].audio_url_answer:"";
      if(target.audio_url) {
        this.g_cartoon.playAudio(target.audio_url, false, ()=>{
          this.g_enableMapDown = true;
        })
      } else {
        this.g_enableMapDown = true;
      }
    })

    line.show = () => {
      tweenChange(mask.ref, {x: this.g_canvasWidth}, 1)
    }

    parent.addChild(line.label.ref);
    parent.addChild(mask.ref);
    parent.addChild(target.ref);
    return line
  }

  nextQuestion(callback?) {
    this.m_currentQuestionIndex++;
    if(this.m_currentQuestionIndex >= this.g_formData.dataArray.length) {
      callback && callback(false)
    } else {
      callback && callback(true)
    }
  }

  createTestLabel() {
    let testLabel_1 = this.g_cartoon.createCartoonElementLabel(`test-label-1`, "TEST_YOU_CANT_SEE_THIS", "BRLNSDB_1", "#FFFFFF", 42 * this.g_mapScale);
    let testLabel_2 = this.g_cartoon.createCartoonElementLabel(`test-label-2`, "TEST_YOU_CANT_SEE_THIS", "DFChu-W4-WIN-BF", "#FFFFFF", 42 * this.g_mapScale);
    let testLabel_3 = this.g_cartoon.createCartoonElementLabel(`test-label-3`, "TEST_YOU_CANT_SEE_THIS", "FuturaStd-Medium", "#FFFFFF", 42 * this.g_mapScale);
    let testLabel_4 = this.g_cartoon.createCartoonElementLabel(`test-label-4`, "TEST_YOU_CANT_SEE_THIS", "GOTHICB_1", "#FFFFFF", 42 * this.g_mapScale);
    this.render(testLabel_1.ref)
    this.render(testLabel_2.ref)
    this.render(testLabel_3.ref)
    this.render(testLabel_4.ref)
  }

































  // --------------------------------------------------
  // --------------  Template function  ---------------
  // --------------------------------------------------
  // --------------------------------------------------
  // --------------------------------------------------
  // 
  //                                  _________
  //                                 /         /.
  //           .-------------.      /_________/ |
  //         /             / |      |         | |
  //        /+============+\ |      | |====|  | |
  //        ||C:\>        || |      |         | |
  //        ||            || |      | |====|  | |
  //        ||            || |      |   ___   | |
  //        ||            || |      |  |166|  | |
  //        ||            ||/@@@    |   ---   | |
  //        \+============+/    @   |_________|./.
  //                          @          ..  ....'
  //        ..................@     __.'.'  ''
  //      /oooooooooooooooo//     ///
  //      /................//     /_/
  //      ------------------
  //
  // --------------------------------------------------

  @ViewChild('canvas', {static: true }) canvas: ElementRef;
  @ViewChild('wrap', {static: true }) wrap: ElementRef;
  @HostListener("window:resize", ["$event"])
  onResize(event) {
    this.g_winResizeEventStream.next();
  }
  ngOnDestroy() {
    window["curCtx"] = null;
    this.g_cartoon.stopAllAudio()
    window.cancelAnimationFrame(this.g_animationId);
  }
  ngOnInit() {
    const getData = (<any>window).courseware.getData;
    getData(data => {
      // if (window['air'].airClassInfo.user.classRole == 'tea') {
      //   if(!this.g_ForceChangeDefaultRole){
      //     this.g_teacherFlag = true;
      //   }
      // }
      // this.g_currentUser = window['air'].airClassInfo.user;
      if (data && typeof data == "object") {
        this.g_data = data;
        this.g_formData = data.contentObj;
      } else {
        this.g_data = {};
      }

      
      if (!this.g_data.contentObj) {
        this.g_data.contentObj = {};
        this.g_formData = {};
      }

      this.initDefaultData();
      this.initAudio();
      this.initImg();

       // 预加载资源
       this.g_cartoon.loadResources().then(() => {
        window["air"].hideAirClassLoading(this.g_KEY, this.g_data);
        this.initSystem();
        this.update();
        this.systemReady()
      });

      this.initListener();
    }, this.g_KEY);
  }

  // ----------------------------------
  // 初始化默认数据
  // ----------------------------------
  initDefaultData() {
    if ( Object.keys(this.g_formData).length===0 || this.g_formData.version != defaultData.version ) {
      this.g_formData = defaultData;
      console.log("Use default data.", this.g_formData)
    }
  }


  // ----------------------------------
  // 初始化音乐
  // ----------------------------------
  initAudio() {
    const contentObj = this.g_formData;
    if (!contentObj) {
      return;
    }
    // 添加用户上传音效
    let images:Array<string> = this.mapToAduioArray(contentObj)
    images.forEach(image => {
      this.g_cartoon.addAudio( image, image );
    });

    // 添加本地音效
    for( var key in localAudios ){
      this.g_cartoon.addAudio( key, localAudios[key] );
    }
  }

  // ----------------------------------
  // 初始化图片
  // ----------------------------------
  initImg() {
    const contentObj = this.g_formData;
    if (contentObj) {
      const addPicUrl = url => {
        if (url) {
          this.g_cartoon.addImage(url, url);
        }
      };
      let images:Array<string> = this.mapToImageArray(contentObj)
      images.forEach(image => {
        addPicUrl(image)
      });
    }
    // 添加本地图片
    for( var key in localImages ){
     this.g_cartoon.addImage( key, localImages[key] );
    }
  }


  mapDown(event) {
    let myStopPropagation = false;
    if (!this.g_enableMapDown) {
      return;
    }

    this.m_mapDownArray.forEach((item)=>{
      if(!myStopPropagation){
        if (this.checkClickTarget(this.g_cartoon.getCartoonElementRef(item.id))) {
          this.g_enableMapDown = false;
          if(this.m_mapDragObject[item.id]){
            this.m_mapDragObject[item.id].trigger()
            let dragElement = this.g_cartoon.getCartoonElement(item.id)
            if(dragElement){
              this.g_currentDragElementID = item.id
              this.g_dragFirstClickX = this.g_clickX
              this.g_dragFirstClickY = this.g_clickY
              this.enableMoveAsstant(()=>{
                this.m_mapDragObject[item.id].move(dragElement)
              })
            }
            myStopPropagation = true;
          }else{
            if(item.callback()){
              myStopPropagation = true;
            }
          }
        }
      }
    })
  }

  mapMove(event) {
    let myStopPropagation = false;
    if (!this.g_enableMapMove) {
      return;
    }

    this.m_mapMoveArray.forEach((item)=>{
      if(!myStopPropagation){
        if (this.checkClickTarget(this.g_cartoon.getCartoonElementRef(item.id))) {
          this.g_enableMapMove = false;
          if(item.callback()){
            myStopPropagation = true;
          }
        }
      }
    })
  }

  mapUp(event) {
    let myStopPropagation = false;
    if (!this.g_enableMapUp) {
      return;
    }
    this.m_mapUpArray.forEach((item)=>{
      if(!myStopPropagation){
        if (this.checkClickTarget(this.g_cartoon.getCartoonElementRef(item.id))) {
          this.g_enableMapUp = false;
          if(item.callback()){
            myStopPropagation = true;
          }
          this.g_currentDragElementID = null;
          this.disableMoveAsstant();
        }
      }
    })
  }

  update() {
    this.g_animationId = window.requestAnimationFrame(this.update.bind(this));
    // 清除画布内容
    this.g_ctx.clearRect(0, 0, this.g_canvasWidth, this.g_canvasHeight);
    TWEEN.update();
    this.updateArr(this.m_renderArr);
    this.updateArr(this.m_endPageArr);
    this.updateArr(this.m_elementPetalArr);
    this.updateArr(this.m_frontCurtainArr)
  }

  updateItem(item) {
    if (item) {
      item.update();
    }
  }

  updateArr(arr) {
    if (!arr) {
      return;
    }
    for (let i = 0; i < arr.length; i++) {
      arr[i].update(this);
    }
  }

  initListener() {
    const element = this.canvas.nativeElement;
    this.g_winResizeEventStream.pipe(debounceTime(500)).subscribe(data => {
      this.renderAfterResize();
    });
    
    const addTouchListener = () => {
      element.addEventListener('touchstart', touchDownFunc);
      element.addEventListener('touchmove', touchMoveFunc);
      element.addEventListener('touchend', touchUpFunc);
      element.addEventListener('touchcancel', touchUpFunc);
    };
    const removeTouchListener = () => {
      element.removeEventListener('touchstart', touchDownFunc);
      element.removeEventListener('touchmove', touchMoveFunc);
      element.removeEventListener('touchend', touchUpFunc);
      element.removeEventListener('touchcancel', touchUpFunc);
    };

    const addMouseListener = () => {
      element.addEventListener('mousedown', mouseDownFunc);
      element.addEventListener('mousemove', mouseMoveFunc);
      element.addEventListener('mouseup', mouseUpFunc);
    };
    const removeMouseListener = () => {
      element.removeEventListener('mousedown', mouseDownFunc);
      element.removeEventListener('mousemove', mouseMoveFunc);
      element.removeEventListener('mouseup', mouseUpFunc);
    };

    const touchDownFunc = (e) => {
      if (this.g_firstTouch) {
        this.g_firstTouch = false;
        removeMouseListener();
      }
      setMxMyByTouch(e);
      this.mapDown(e);
    };
    const touchMoveFunc = (e) => {
      setMxMyByTouch(e);
      this.mapMove(e);
    };
    const touchUpFunc = (e) => {
      setMxMyByTouch(e);
      this.mapUp(e);
    };

    const mouseDownFunc = (e) => {
      if (this.g_firstTouch) {
        this.g_firstTouch = false;
        removeTouchListener();
      }
      setMxMyByMouse(e);
      this.mapDown(e);
    };
    const mouseMoveFunc = (e) => {
      setMxMyByMouse(e);
      this.mapMove(e);
    };
    const mouseUpFunc = (e) => {
      setMxMyByMouse(e);
      this.mapUp(e);
    };

    const setMxMyByTouch = event => {
      if (event.touches.length <= 0) {
        return;
      }
      if (this.g_canvasLeft == null) {
        setParentOffset();
      }
      this.g_clickX = event.touches[0].pageX - this.g_canvasLeft;
      this.g_clickY = event.touches[0].pageY - this.g_canvasTop;
    };

    const setParentOffset = () => {
      const rect = this.canvas.nativeElement.getBoundingClientRect();
      this.g_canvasLeft = rect.left;
      this.g_canvasTop = rect.top;
    };

    const setMxMyByMouse = (event) => {
      this.g_clickX = event.offsetX;
      this.g_clickY = event.offsetY;
    };

    addMouseListener();
    addTouchListener();
  }

  showArr(arr) {
    if (!arr) {
      return;
    }
    for (let i = 0; i < arr.length; i++) {
      arr[i].visible = true;
    }
  }

  hideArr(arr) {
    if (!arr) {
      return;
    }
    for (let i = 0; i < arr.length; i++) {
      arr[i].visible = false;
    }
  }

  IsPC() {
    if (window["ELECTRON"]) {
      return false; // 封装客户端标记
    }
    if (
      document.body.ontouchmove !== undefined &&
      document.body.ontouchmove !== undefined
    ) {
      return false;
    } else {
      return true;
    }
  }

  renderAfterResize() {
    this.g_canvasWidth = this.wrap.nativeElement.clientWidth;
    this.g_canvasHeight = this.wrap.nativeElement.clientHeight;
    this.update();
    this.handleScreenResize()
  }

  checkClickTarget(target) {
    if (!target) {
      return false;
    }
    const rect = target.getBoundingBox();
    if (this.checkPointInRect(this.g_clickX, this.g_clickY, rect)) {
      return true;
    }
    return false;
  }

  getWorlRect(target) {
    let rect = target.getBoundingBox();
    if (target.parent) {
      const pRect = this.getWorlRect(target.parent);
      rect.x += pRect.x;
      rect.y += pRect.y;
    }
    return rect;
  }

  checkPointInRect(x, y, rect) {
    if (x >= rect.x && x <= rect.x + rect.width) {
      if (y >= rect.y && y <= rect.y + rect.height) {
        return true;
      }
    }
    return false;
  }

  getPosByAngle(angle, len) {
    const radian = (angle * Math.PI) / 180;
    const x = Math.sin(radian) * len;
    const y = Math.cos(radian) * len;
    return { x, y };
  }

  getPosDistance(sx, sy, ex, ey) {
    const _x = ex - sx;
    const _y = ey - sy;
    const len = Math.sqrt(Math.pow(_x, 2) + Math.pow(_y, 2));
    return len;
  }

  subscribeMapDownEvent(id,callback, zIndex?){
    zIndex = zIndex?zIndex:0
    this.m_mapDownObject.push({id:id, zIndex:zIndex?zIndex:1, callback:callback})
    this.m_mapDownObject.sort((a,b)=>{
      return b.zIndex-a.zIndex
    })
    this.m_mapDownArray = []
    this.m_mapDownObject.forEach(item=>{
      this.m_mapDownArray.push(item)
    })
  }

  subscribeMapUpEvent(id,callback, zIndex?){
    zIndex = zIndex?zIndex:0
    this.m_mapUpObject.push({id:id, zIndex:zIndex?zIndex:1, callback:callback})
    this.m_mapUpObject.sort((a,b)=>{
      return b.zIndex-a.zIndex
    })
    this.m_mapUpArray = []
    this.m_mapUpObject.forEach(item=>{
      this.m_mapUpArray.push(item)
    })
  }

  subscribeMapMoveEvent(id, callback, zIndex?){
    zIndex = zIndex?zIndex:0
    this.m_mapMoveObject.push({id:id, zIndex:zIndex?zIndex:1, callback:callback})
    this.m_mapMoveObject.sort((a,b)=>{
      return a.zIndex - b.zIndex
    })
    this.m_mapMoveArray = []
    this.m_mapMoveObject.forEach(item=>{
      this.m_mapMoveArray.push(item)
    })
  }

  subscribeMapDragEvent(id, trigger, move, release){
    this.subscribeMapDownEvent(id,()=>{})
    this.m_mapDragObject[id] = {
      trigger,
      move,
      release
    };
  }

  releaseDragElement(){
    if(this.g_currentDragElementID){
      this.m_mapDragObject[this.g_currentDragElementID].release()
    }
  }

  // 全局初始化入口，当资源加载完毕后执行
  initSystem() {
    
    this.g_canvasWidth = this.wrap.nativeElement.clientWidth;
    this.g_canvasHeight = this.wrap.nativeElement.clientHeight;

    const sx = this.g_canvasWidth / this.g_canvasBaseW;
    const sy = this.g_canvasHeight / this.g_canvasBaseH;
    const s = Math.min(sx, sy);

    this.g_mapScale = s;
    this.g_cartoon.mapScale = this.g_mapScale;
    this.g_cartoon.clientWidth = this.wrap.nativeElement.clientWidth;
    this.g_cartoon.clientHeight = this.wrap.nativeElement.clientHeight;

    this.m_renderArr = [];
    this.m_frontCurtainArr = [];
    this.m_renderObject = [];
    this.g_enableMapDown = true;

    this.g_ctx = this.canvas.nativeElement.getContext("2d");
    this.canvas.nativeElement.width = this.g_canvasWidth;
    this.canvas.nativeElement.height = this.g_canvasHeight;
    window["curCtx"] = this.g_ctx;

    // 初始化舞台
    this.initStage();
    this.update();
  }
  

  initStage() {
    const bgWidth = 1280;
    const bgHeight = 720;
    this.g_cartoon.stageWidth = bgWidth*this.g_mapScale;
    this.g_cartoon.stageHeight= bgHeight*this.g_mapScale;

    const imageColor = this.g_cartoon.createCartoonElement("g-ackground-color", "ShapeRect").ref;
    // console.log( "stageWidth: " + this.g_cartoon.stageWidth + " stageHeight" + this.g_cartoon.stageHeight)
    // console.log( "clientWidth: " + this.g_cartoon.clientWidth + " clientHeight" + this.g_cartoon.clientHeight)
    // console.log( "canvasWidth: " + this.g_canvasWidth + " canvasHeight" + this.g_canvasHeight)
    // console.log( "mapScale: " + this.g_mapScale )
    imageColor.x = this.g_cartoon.clientWidth / 2 ;
    imageColor.y = this.g_cartoon.clientHeight / 2 ;
    imageColor.setSize(this.g_cartoon.clientWidth, this.g_cartoon.clientHeight);
    imageColor.init();
    imageColor.fillColor = this.g_background_color;
    this.render(imageColor, -999999)

    const image = this.g_cartoon.createCartoonElement("bg_1280_720_Ruler", "MySprite").ref;
    image.init(this.g_cartoon.images.get("_bg_1280_720_Ruler"));
    image.x = this.g_canvasWidth / 2;
    image.y = this.g_canvasHeight /2;
    image.visible = this.g_EnableStageRuler
    this.g_cartoon.setOrigin( image.x-(bgWidth/2*this.g_mapScale), image.y-(bgHeight/2*this.g_mapScale)  )
    this.g_cartoon.setRelativeOrigin( -bgWidth/2, -bgHeight/2)
    image.setScaleXY(this.g_mapScale);
    this.render(image, -999)
    this.g_stage = image

    const topElementContainer = this.g_cartoon.createCartoonElement("_bg_1280_720", "MySprite").ref;
    topElementContainer.init(this.g_cartoon.images.get("_bg_1280_720"));
    topElementContainer.x = this.g_canvasWidth / 2;
    topElementContainer.y = this.g_canvasHeight /2;
    topElementContainer.setScaleXY(this.g_mapScale);
    this.render(topElementContainer, 9999)
    this.g_topElementContainer = topElementContainer
    
    const maskLayer = this.g_cartoon.createCartoonElement("_mask_layer_1280_720", "MySprite").ref;
    maskLayer.init(this.g_cartoon.images.get("_mask_layer_1280_720"));
    maskLayer.scaleX = this.g_canvasWidth / maskLayer.width;
    maskLayer.scaleY = this.g_canvasHeight / maskLayer.height;
    maskLayer.alpha = 0
    maskLayer.x = this.g_canvasWidth / 2;
    maskLayer.y = this.g_canvasHeight /2;
    this.render(maskLayer, this.g_maskLayerZindex)

    this.subscribeMapUpEvent("g-ackground-color", ()=>{
      if(this.g_currentDragElementID){
        this.releaseDragElement()
        this.g_currentDragElementID = null;
      }else{
        this.g_enableMapUp = true;
      }
    })
  }

  render(ele, zIndex?:number){
    this.m_renderObject.push({element:ele, zIndex:zIndex?zIndex:1})
    this.m_renderObject.sort((a,b)=>{
      return a.zIndex - b.zIndex
    })
    this.m_renderArr = []
    this.m_renderObject.forEach(item=>{
      this.m_renderArr.push(item.element)
    })    
  }

  sendServerEvent(key, data) {
    const c = (<any> window).courseware;
    c.sendEvent(key, JSON.stringify(data));
  }

  addServerListener(msg_key, callback) {
    const c = (<any> window).courseware;
    c.onEvent(msg_key, (data,next) => {
      callback(JSON.parse(data))
      next && next();
    });
  }

  randomArray_shuffle(array) {
    var input = array;
    for (var i = input.length-1; i >=0; i--) {
        var randomIndex = Math.floor(Math.random()*(i+1));
        var itemAtIndex = input[randomIndex];
        input[randomIndex] = input[i];
        input[i] = itemAtIndex;
    }
    return input;
  }

  paginationArray(pageNo, pageSize, array) {
    var offset = (pageNo - 1) * pageSize;
    return (offset + pageSize >= array.length) ? array.slice(offset, array.length) : array.slice(offset, offset + pageSize);
  }

  stopAllTimeout(){
    this.m_setTimeoutIDs.forEach(id=>clearTimeout(id))
    this.m_setTimeoutIDs = []
  }

  stopAllInterval(){
    this.m_setIntervalIDs.forEach(id=>clearInterval(id))
    this.m_setIntervalIDs = []
  }

  topOfRenderArray(element){
    let index = this.m_renderArr.indexOf(element)
    if(index !=-1){
      this.m_renderArr.splice(index, 1)
      this.m_renderArr.push(element)
    }
  }

  setRenderZIndex(element, zIndex){
    let index = null;
    for(let i=0; i<this.m_renderObject.length; i++){
      if( this.m_renderObject[i].element.id == element.id ){
        index = i
      }
    }
    if(index){
      this.m_renderObject.splice(index, 1)
      this.m_renderObject.push({element:element, zIndex:zIndex?zIndex:1})
      this.m_renderObject.sort((a,b)=>{
        return a.zIndex - b.zIndex
      })
  
      this.m_renderArr = []
      this.m_renderObject.forEach(item=>{
        this.m_renderArr.push(item.element)
      })
    }
  }
  
  deleteElementInRender(id){
    let index = null;
    for(let i=0; i<this.m_renderObject.length; i++){
      if( this.m_renderObject[i].element.id == id ){
        index = i
      }
    }
    if(index){
      this.m_renderObject.splice(index, 1)
      this.m_renderObject.sort((a,b)=>{
        return a.zIndex - b.zIndex
      })
      this.m_renderArr = []
      this.m_renderObject.forEach(item=>{
        this.m_renderArr.push(item.element)
      })
    }else{
      console.warn("Can not found element id:" + id)
    }
  }

  enableMoveAsstant(callback){
    if (this.m_moveAsstantIntervalId) {
      this.g_enableDragging = false;
      clearInterval(this.m_moveAsstantIntervalId)
    }
    this.m_moveAsstantIntervalId = setInterval(() => {
      if(this.g_enableDragging){
        callback()
      }else{
        let moveX = Math.abs(this.g_dragFirstClickX - this.g_clickX)
        let moveY = Math.abs(this.g_dragFirstClickY - this.g_clickY)
        if (moveX > this.g_enableDragDistance || moveY > this.g_enableDragDistance) {
          this.g_enableDragging = true;
        }
      }
    }, 50)
  }

  disableMoveAsstant(){
    this.g_enableDragging = false;
    clearInterval(this.m_moveAsstantIntervalId)
  }

  cleanSystemVar(){
    this.m_mapDownQueue = {}
    this.m_mapDownArray = []
    this.m_mapDownObject = []
    this.m_mapMoveArray = []
    this.m_mapMoveObject = []
    this.m_mapUpQueue = {}
    this.m_mapUpArray = []
    this.m_mapUpObject = []
    this.m_mapDragQueue = []
    this.m_mapDragArray = []
    this.m_mapDragObject = []
    this.stopAllInterval();
    this.stopAllTimeout()
    this.g_cartoon.stopAllAudio()
  }

  getMaxSubstringLength(str){
    let maxLength = 0;
    let subSubstring = str.split(" ")
    for (let index=0; index<subSubstring.length; index++) {
      if(subSubstring[index].length > maxLength){
        maxLength = subSubstring[index].length;
      }
    }
    return maxLength
  }

  setLeftCornerTest(){
    const bgRect = new ShapeRect();
    bgRect.setSize(57, 65);
    bgRect.fillColor = '#f8c224';
    const sx = this.g_canvasWidth / this.g_canvasBaseW;
    bgRect.setScaleXY(sx);
    bgRect.x = 65 * sx;
    this.g_partTitle_x = bgRect.x
    this.g_mainTitle_x = bgRect.x + 80 * sx
    bgRect.alpha = 0.5
    bgRect.visible = this.g_showLeftCornerTest
    this.render(bgRect, 9999);
  }

  alignCenter(elementArray, spacing, laseX = 0, withAni?, callback?){
    let totlaWidth = 0
    let length = elementArray.length
    elementArray.forEach(element => {
      let bd = element.ref.getBoundingBox()
      totlaWidth += bd.width
    });
    totlaWidth += (elementArray.length-1) * spacing
    // laseX = laseX? laseX: this.g_canvasWidth / 2 - totlaWidth/2;
    elementArray.forEach((element, index) => {
      let bd = element.ref.getBoundingBox()
      if(withAni){
        tweenChange(element.ref, {x: laseX + bd.width / 2}, 0.2, ()=>{
          this.g_cartoon.saveSize(element.id)
          if(length == (index+1)){
            callback && callback()
          }
        })
      }else{
        element.ref.x = laseX + bd.width / 2
        this.g_cartoon.saveSize(element.id)
        if(length == (index+1)){
          callback && callback()
        }
      }

      laseX = (laseX + bd.width + spacing)
    });
    return totlaWidth
  }

  alignLeft(elementArray, spacing, startX = 0, withAni?, callback?){
    let laseX = startX;
    elementArray.forEach((element, index) => {
      let bd = element.ref.getBoundingBox()
      if(withAni){
        tweenChange(element.ref, {x: laseX + bd.width / 2}, 0.2, ()=>{
          this.g_cartoon.saveSize(element.id)
          if(length == (index+1)){
            callback && callback()
          }
        })
      }else{
        element.ref.x = laseX + bd.width / 2
        this.g_cartoon.saveSize(element.id)
        if(length == (index+1)){
          callback && callback()
        }
      }

      laseX = (laseX + bd.width + spacing)
    });
  }

  createTestLine(x, y, height, color?){
    var colorAll = ['#ff0000','#eb4310','#f6941d','#fbb417','#ffff00','#cdd541','#99cc33','#3f9337','#219167','#239676','#24998d','#1f9baa','#0080ff','#3366cc','#333399','#003366','#800080','#a1488e','#c71585','#bd2158'];
    color = color ? color : colorAll[Math.floor(Math.random()*20)]
    this.render(this.g_cartoon.createRectangula({
      width: 5,
      height: height,
      x: x,
      y: y,
      fillColor: color
    }), 9999)
  }

  getSuitSizeBackground(key, width){
    let bgAll = multiSizeBackground[key]
    if(!bgAll){
      return ""
    }
    let suitFileName = ""
    for(let fileName in bgAll){
      if(  Number(fileName) >= width){
        suitFileName = bgAll[fileName]
        break
      }
    }
    return suitFileName
  }












  













  












  // --------------------------------------------------
  // --------------  Template function  ---------------
  // --------------------------------------------------
  // --------------------------------------------------
  // --------------------------------------------------
  // 
  //                .-~~~~~~~~~-._       _.-~~~~~~~~~-.
  //            __.'              ~.   .~              `.__
  //          .'//                  \./                  \\`.
  //        .'//                     |                     \\`.
  //      .'// .-~"""""""~~~~-._     |     _,-~~~~"""""""~-. \\`.
  //    .'//.-"                 `-.  |  .-'                 "-.\\`.
  //  .'//______.============-..   \ | /   ..-============.______\\`.
  //.'______________________________\|/______________________________`.
  //
  // --------------------------------------------------
  // showParticle( element_id )                 泡泡效果
  // --------------------------------------------------
  // showEndPatal()  /  stopEndPatal()  花瓣飘落结束动画 
  // --------------------------------------------------
  // showCorrectPatal()                 指定元素上面飘花
  // --------------------------------------------------
  // convertPercentToRadian()   将百分比转换为弧长 第一个参数是百分比，第二个参数是方向 true为逆时针，false为顺时针。用于ShapeCircle换圆弧
  // --------------------------------------------------
  // movePaoWuxian()                        元素抛物线跳
  // --------------------------------------------------
  // showJellyAnimation()------------------------------
  // showBlingStar()----------------------显示星星效果--
  // --------------------------------------------------
  


  maskLayerShow(isShow, time) {
    let element = this.g_cartoon.getCartoonElement("_mask_layer_1280_720")
    if(isShow) {
      tweenChange(element.ref, {alpha: 1}, time)
    } else {
      tweenChange(element.ref, {alpha: 0}, time)
    }
  }

  // 泡泡
  showParticle(card) {
    let myCard = this.g_cartoon.getCartoonElementRelativePosition(card.id)
    showPopParticle(this.g_cartoon.images.get("_bubble"), { x: myCard.x , y: myCard.y }, this.g_stage);
  }

  // 选择正确动画
  showCorrectPatal(card_id, showTime, callback?) {
    this.m_elementPetalArr = [];
    this.m_showElementPetalFlag = true;
    this.addCorrectPetal(card_id);
    setTimeout(()=>{
      this.m_elementPetalArr = [];
      this.m_showElementPetalFlag = false;
      callback && callback()
    },showTime)
  }
  stopAllCorrectPatal(){
    this.m_elementPetalArr = [];
    this.m_showElementPetalFlag = false;
  }

  addCorrectPetal(card_id) {
    if (!this.m_showElementPetalFlag) {
      return;
    }

    let element = this.g_cartoon.getCartoonElement(card_id)
    const petal = new MySprite(this.g_ctx);

    const id = Math.ceil(Math.random() * 3);
    petal.init(this.g_cartoon.images.get(this.m_PetalImage + id));

    const randomS = (Math.random() * 0.4 + 0.6) * this.g_mapScale * 0.5;
    petal.setScaleXY(randomS);

    const randomR = Math.random() * 360;
    petal.rotation = randomR;

    const randomX = Math.random() * element.ref.width * this.g_mapScale;

    petal.x = element.ref.x - (element.ref.width * this.g_mapScale / 2) + randomX;
    petal.y = element.ref.y - (element.ref.height * this.g_mapScale / 2);

    const randomT = 1 + Math.random() * 2;
    petal["time"] = randomT;

    let randomTR = 360 * Math.random(); // - 180;
    if (Math.random() < 0.5) {
      randomTR *= -1;
    }
    petal["tr"] = randomTR;

    this.m_elementPetalArr.push(petal);
    moveItem(
      petal,
      petal.x,
      element.ref.y + (element.ref.height * this.g_mapScale / 2),
      petal["time"],
      () => {
        removeItemFromArr(this.m_elementPetalArr, petal);
      }
    );
    rotateItem(petal, petal["tr"], petal["time"]);
    setTimeout(() => {
      this.addCorrectPetal(card_id);
    }, 200);
  }


  // 结束动画花瓣飘落
  showEndPatal() {
    this.m_endPageArr = [];
    this.m_showPetalFlag = true;
    this.addPetal();
  }

  stopEndPatal() {
    this.m_endPageArr = [];
    this.m_showPetalFlag = false;
  }

  addPetal() {
    if (!this.m_showPetalFlag) {
      return;
    }
    const petal = this.getPetal();
    this.m_endPageArr.push(petal);
    moveItem(
      petal,
      petal.x,
      this.g_canvasHeight + petal.height * petal.scaleY,
      petal["time"],
      () => {
        removeItemFromArr(this.m_endPageArr, petal);
      }
    );
    rotateItem(petal, petal["tr"], petal["time"]);
    setTimeout(() => {
      this.addPetal();
    }, 100);
  }

  getPetal() {
    const petal = new MySprite(this.g_ctx);

    const id = Math.ceil(Math.random() * 3);
    petal.init(this.g_cartoon.images.get(this.m_PetalImage + id));

    const randomS = (Math.random() * 0.4 + 0.6) * this.g_mapScale;
    petal.setScaleXY(randomS);

    const randomR = Math.random() * 360;
    petal.rotation = randomR;

    const randomX = Math.random() * this.g_canvasWidth;
    petal.x = randomX;
    petal.y = (-petal.height / 2) * petal.scaleY;

    const randomT = 2 + Math.random() * 5;
    petal["time"] = randomT;

    let randomTR = 360 * Math.random(); // - 180;
    if (Math.random() < 0.5) {
      randomTR *= -1;
    }
    petal["tr"] = randomTR;

    return petal;
  }


  showJellyAnimation(element_id, callback?){
    let element = this.g_cartoon.getCartoonElement(element_id).ref
    tweenChange(element,{ scaleX: this.g_mapScale * element.initScaleX * 1.1 , scaleY: this.g_mapScale * element.initScaleY * 1.1 }, 0.1, ()=>{
      tweenChange(element,{ scaleX: this.g_mapScale * element.initScaleX * 0.9 , scaleY: this.g_mapScale * element.initScaleY * 0.9 }, 0.1, ()=>{
        tweenChange(element,{ scaleX: element.initScaleX , scaleY: element.initScaleY }, 0.1, ()=>{
          callback && callback()
        })
      })
    })
  }

  showShakeAnimation(element_id, callback?){
    let element = this.g_cartoon.getCartoonElement(element_id).ref
    let originX = element.x
    tweenChange(element,{ x: originX - 10*this.g_mapScale }, 0.1, ()=>{
      tweenChange(element,{ x: originX + 10*this.g_mapScale }, 0.1, ()=>{
        tweenChange(element,{ x: originX }, 0.1, ()=>{
          callback && callback()
        })
      })
    })
  }

  convertPercentToRadian(per, direction){
    let radian = 0;
    if(direction) { //逆时针
      if(per*360 > 270){
        radian = 361 - (per*360 - 270)
      }else{
        radian = 270 - per*360
      }
    }else{ //顺时针
      if(per*360 > 90){
        radian = per*360 - 90
      }else{
        radian = 270 + per*360
      }
    }
    return (radian * Math.PI) / 180;
  }

  showBlingStar(element, callback?){
    let rect =  element.getBoundingBox()
    showBlingBling(this.g_cartoon.images.get('icon_star'), rect, element, 0.5, 1, 0.08, 0.5);
    this.m_setTimeoutIDs.push(setTimeout(() => {
      callback && callback()
    }, 2000))
    // setTimeout(()=>{
    //   showBlingBling(this.g_cartoon.images.get('icon_star'), rect, element, 0.5, 1, 0.08, 1.8);
    // },200)
    // setTimeout(()=>{
    //   showBlingBling(this.g_cartoon.images.get('icon_star'), rect, element, 0.5, 1, 0.08, 1.6);
    // },400)
  }

  movePaoWuxian(element,dis,time,callback?){
    let disX = dis.x - element.x
    let disY = (element.y - dis.y) 
    let count = 0
    let runTime = time/5
    let vx = disX/runTime
    let a = -0.05
    let vy0 = disY/runTime - 0.5*a*runTime
    let id = setInterval(()=>{
      element.x += vx
      element.y -= vy0 + a*count
      count++;
      if(count>runTime){
        clearInterval(id)
        callback && callback()
      }
    }, 5);
  }

  showBalloons(maxBalloon = 10) {
    let balloonImgList = [
      "_balloon_blue",
      "_balloon_green",
      "_balloon_orange",
      "_balloon_pink",
      "_balloon_red",
      "_balloon_yellow",
    ];
  
    for (var i = 0; i < maxBalloon; ++i) {
      let balloon = new MySprite();
      balloon.init(
        this.g_cartoon.images.get(balloonImgList[this.randomInt(balloonImgList.length)])
      );
      let leftBorder = -640;
      let rightBorder = 640;
      balloon.x = this.randomInt(leftBorder, rightBorder);
      balloon.y = 720 / 2 + this.randomInt(400);
      this.g_topElementContainer.addChild(balloon);
  
      balloon.scaleX = 0;
      balloon.scaleY = 0;
      tweenChange(
        balloon, {
          scaleX: 1 + Math.random() * 0.3 - 0.15,
          scaleY: 1 + Math.random() * 0.3 - 0.15,
        },
        0.5
      );
  
      let wholeTime = 3;
      if (i < 4) {
        wholeTime = 3 + i;
      }
  
      tweenChange(
        balloon,
        { y: -720 / 2 - this.randomInt(200) },
        wholeTime,
        ()=>{
          this.g_topElementContainer.removeChild(balloon)
        },
        TWEEN.Easing.Quadratic.In
      );
  
      let time = Math.random() * wholeTime;
      tweenChange(
        balloon,
        {
          x: this.Between(
            leftBorder,
            rightBorder,
            balloon.x + Math.min(time * 50, this.randomInt(200))
          ),
        },
        time,
        () => {
          tweenChange(
            balloon,
            {
              x: this.Between(
                leftBorder,
                rightBorder,
                balloon.x - Math.min(time * 50, this.randomInt(200))
              ),
            },
            wholeTime - time
          );
        }
      );
    }
  }

  randomInt(a, b = 0) {
    let max = Math.max(a, b);
    let min = Math.min(a, b);
    return Math.floor(Math.random() * (max - min) + min);
  }

  Between(a, b, c) {
    return [a, b, c].sort((a, b) => a - b)[1];
  }

}